package it.cavallium.strangedb.database.io;


import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.file.Path;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class ConcurrentMultiFileIO implements IMultiFileIO {

	public final int filesCount;
	private final SimpleFileReaderWriter[] fileRW;
	private ReentrantReadWriteLock[] locks;

	public ConcurrentMultiFileIO(Path dataFile, int filesCount) throws IOException {
		this.filesCount = filesCount;
		String fileName = dataFile.toFile().getName();

		fileRW = new SimpleFileReaderWriter[filesCount];
		locks = new ReentrantReadWriteLock[filesCount];

		for (int i = 0; i < filesCount; i++) {
			locks[i] = new ReentrantReadWriteLock(false);
			fileRW[i] = new SimpleFileReaderWriter(dataFile.resolveSibling(ExtensionUtils.insertExtension(fileName, i)));
		}
	}

	@Override
	public int getFilesCount() {
		return filesCount;
	}

	@Override
	public ByteBuffer readAt(int fileIndex, long index, int length) throws IOException {
		locks[fileIndex].readLock().lock();
		try {
			return fileRW[fileIndex].readAt(index, length);
		} finally {
			locks[fileIndex].readLock().unlock();
		}
	}

	@Override
	public int writeAt(int fileIndex, long index, int length, ByteBuffer data) throws IOException {
		locks[fileIndex].writeLock().lock();
		try {
			return fileRW[fileIndex].writeAt(index, length, data);
		} finally {
			locks[fileIndex].writeLock().unlock();
		}
	}

	@Override
	public long writeAtEnd(int fileIndex, int length, ByteBuffer data) throws IOException {
		locks[fileIndex].writeLock().lock();
		try {
			return fileRW[fileIndex].writeAtEnd(length, data);
		} finally {
			locks[fileIndex].writeLock().unlock();
		}
	}

	@Override
	public void close() throws IOException {
		for (ReentrantReadWriteLock lock : locks) {
			lock.writeLock().lock();
		}
		try {
			for (SimpleFileReaderWriter file : fileRW) {
				file.close();
			}
		} finally {
			for (ReentrantReadWriteLock lock : locks) {
				lock.writeLock().unlock();
			}
		}
	}

	public long getTotalSize() {
		long totalSize = 0;
		for (int i = 0; i < fileRW.length; i++) {
			locks[i].readLock().lock();
			totalSize += fileRW[i].getSize();
			locks[i].readLock().unlock();
		}
		return totalSize;
	}
}
