package it.cavallium.strangedb.database.references;

import it.cavallium.strangedb.database.io.ConcurrentMultiFileIO;
import it.cavallium.strangedb.database.io.DatabaseIO;
import it.cavallium.strangedb.database.IReferencesIO;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.file.Path;

public class ConcurrentDatabaseReferencesIO implements IReferencesIO {

	private final DatabaseIO fileIO;
	private final ConcurrentDatabaseReferencesMetadata referencesMetadata;

	public ConcurrentDatabaseReferencesIO(Path dataFile, Path referencesMetaFile, int dataFilesCount, int referencesFilesCount) throws IOException {
		this.fileIO = new DatabaseIO(new ConcurrentMultiFileIO(dataFile, dataFilesCount));
		this.referencesMetadata = new ConcurrentDatabaseReferencesMetadata(referencesMetaFile, referencesFilesCount);
	}

	@Override
	public long allocateReference() throws IOException {
		return referencesMetadata.newReference(referencesMetadata.allocateReference(), 0, 0);
	}

	@Override
	public long allocateReference(int size, ByteBuffer data) throws IOException {
		long reference = referencesMetadata.allocateReference();
		long index = writeToFile(reference, size, data);
		return referencesMetadata.newReference(reference, index, size);
	}

	@Override
	public void writeToReference(long reference, byte cleanerId, int size, ByteBuffer data) throws IOException {
		long index = writeToFile(reference, size, data);
		referencesMetadata.editReference(reference, new ReferenceInfo(index, size, cleanerId));
	}

	@Override
	public ByteBuffer readFromReference(long reference) throws IOException {
		ReferenceInfo referenceInfo = referencesMetadata.getReferenceInfo(reference);
		return fileIO.readAt((int) reference, referenceInfo.getIndex(), referenceInfo.getSize());
	}

	@Override
	public void close() throws IOException {
		this.referencesMetadata.close();
		this.fileIO.close();
	}

	public ByteBuffer readFromReferenceSizeAndLastElementOfReferencesList(long reference) throws IOException {
		ReferenceInfo referenceInfo = referencesMetadata.getReferenceInfo(reference);
		if (referenceInfo.getSize() >= Integer.BYTES * 2 + Long.BYTES) {
			return fileIO.readAt((int) reference, referenceInfo.getIndex() + referenceInfo.getSize() - (Integer.BYTES + Long.BYTES), Integer.BYTES + Long.BYTES);
		} else {
			return fileIO.readAt((int) reference, referenceInfo.getIndex(), referenceInfo.getSize());
		}
	}

	/**
	 *
	 * @param size
	 * @param data
	 * @return index
	 * @throws IOException
	 */
	private long writeToFile(long hash, int size, ByteBuffer data) throws IOException {
		if (size == 0 && data == null) return 0;
		if (size < 0) {
			throw new IOException("Trying to create a block with size " + size);
		}
		if (data.limit() < size) {
			throw new IOException("Trying to create a block with size " + size + " but with a buffer of size " + data.limit());
		}
		return fileIO.writeAtEnd((int) hash, size, data);
	}

	public ReferenceInfo getReferenceInfo(long reference) throws IOException {
		return referencesMetadata.getReferenceInfo(reference);
	}

	public long getFirstFreeReference() {
		return referencesMetadata.getFirstFreeReference();
	}
}
