package it.cavallium.strangedb.database.references;

import it.cavallium.strangedb.database.IReferencesMetadata;
import it.cavallium.strangedb.database.io.ConcurrentMultiFileIO;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import static it.cavallium.strangedb.database.IDatabase.DISK_BLOCK_SIZE;

public class ConcurrentDatabaseReferencesMetadata implements IReferencesMetadata {
	public static final byte ERRORED_CLEANER = (byte) -1;
	public static final byte BLANK_DATA_CLEANER = (byte) -2;
	public static final ReferenceInfo NONEXISTENT_REFERENCE_INFO = new ReferenceInfo(-1, -1, ERRORED_CLEANER);
	private static final int REF_META_BYTES_COUNT = Long.BYTES + Integer.BYTES + Byte.BYTES;
	public static final int REF_META_READS_AT_EVERY_READ = (DISK_BLOCK_SIZE - DISK_BLOCK_SIZE % REF_META_BYTES_COUNT) / REF_META_BYTES_COUNT;
	private static final int SEQUENCE_SIZE = 6;
	private static final long SEQUENCE_SIZE_L = SEQUENCE_SIZE;

	private final int filesCount;
	private final long filesCountL;
	private final ConcurrentMultiFileIO metaFile;
	private final DatabaseReferencesMetadataCache cache;
	private AtomicLong firstFreeReference;
	private final ReentrantReadWriteLock lock = new ReentrantReadWriteLock(false);

	public ConcurrentDatabaseReferencesMetadata(Path refMetaFile, int filesCount) throws IOException {
		this.filesCount = filesCount;
		this.filesCountL = filesCount;
		metaFile = new ConcurrentMultiFileIO(refMetaFile, this.filesCount);
		firstFreeReference = new AtomicLong(metaFile.getTotalSize() / REF_META_BYTES_COUNT);
		this.cache = new DatabaseReferencesMetadataCache(filesCount, new Flusher());
	}

	@Override
	public ReferenceInfo getReferenceInfo(long reference) throws IOException {
		ReferenceInfo block;
		long[] allReferences;
		ReferenceInfo[] allInfo;
		lock.readLock().lock();
		try {
			long firstFreeReference = this.firstFreeReference.get();
			if (reference >= firstFreeReference) {
				return NONEXISTENT_REFERENCE_INFO;
			}
			if ((block = cache.getInfo(reference)) != NONEXISTENT_REFERENCE_INFO) {
				return block;
			}
			int size = REF_META_READS_AT_EVERY_READ * REF_META_BYTES_COUNT;
			if (reference + (REF_META_READS_AT_EVERY_READ - 1) >= firstFreeReference) {
				size = (int) ((firstFreeReference - reference)) * REF_META_BYTES_COUNT;
			}
			int referencesCount = size / REF_META_BYTES_COUNT;
			if (referencesCount < 1) {
				throw new IOException("Trying to read <1 references");
			}

			int fileId = toFileId(reference);
			long localReference = toLocalReference(fileId, reference);

			ByteBuffer buffer = metaFile.readAt(fileId, localReference * REF_META_BYTES_COUNT, size);

			if (buffer.limit() % REF_META_BYTES_COUNT != 0 || buffer.limit() < REF_META_BYTES_COUNT) {
				throw new IOException("The buffer is smaller than the data requested.");
			} else if (buffer.limit() != size) {
				size = buffer.limit();
				referencesCount = size / REF_META_BYTES_COUNT;
			}

			allReferences = new long[referencesCount];
			allInfo = new ReferenceInfo[referencesCount];

			block = NONEXISTENT_REFERENCE_INFO;
			for (int delta = 0; delta < referencesCount; delta++) {
				long referenceToLoad = toReference(fileId, localReference + delta);
				long currentIndex = buffer.getLong();
				int currentSize = buffer.getInt();
				byte cleanerId = buffer.get();
				ReferenceInfo refInfo = new ReferenceInfo(currentIndex, currentSize, cleanerId);
				if (buffer.limit() != 0 && currentIndex != 0xFFFFFFFFFFFFFFFFL) {
					allReferences[delta] = referenceToLoad;
					allInfo[delta] = refInfo;
					if (referenceToLoad == reference) {
						block = refInfo;
					}
				} else {
					allReferences[delta] = referenceToLoad;
					allInfo[delta] = refInfo;
					if (referenceToLoad == reference) {
						block = NONEXISTENT_REFERENCE_INFO;
					}
				}
			}
		} finally {
			lock.readLock().unlock();
		}
		lock.writeLock().lock();
		try {
			cache.putAll(allReferences, allInfo);
		} finally {
			lock.writeLock().unlock();
		}
		return block;
	}

	@Override
	public long allocateReference() throws IOException {
		return firstFreeReference.getAndIncrement();
	}

	@Override
	public long newReference(long newlyAllocatedReference, long index, int size) throws IOException {
		cache.put(newlyAllocatedReference, new ReferenceInfo(index, size, BLANK_DATA_CLEANER), true, true);
		return newlyAllocatedReference;
	}

	@Override
	public void editReference(long reference, ReferenceInfo info) throws IOException {
		lock.writeLock().lock();
		try {
			cache.put(reference, info, true, false);
		} finally {
			lock.writeLock().unlock();
		}
	}

	@Override
	public void deleteReference(long reference) throws IOException {
		lock.writeLock().lock();
		try {
			cache.put(reference, NONEXISTENT_REFERENCE_INFO, true, false);
		} finally {
			lock.writeLock().unlock();
		}
	}

	@Override
	public void close() throws IOException {
		lock.writeLock().lock();
		try {
			cache.close();
			metaFile.close();
		} finally {
			lock.writeLock().unlock();
		}
	}

	@Override
	public long getFirstFreeReference() {
		lock.readLock().lock();
		try {
			return firstFreeReference.get();
		} finally {
			lock.readLock().unlock();
		}
	}

	private class Flusher implements DatabaseReferencesMetadataCacheFlusher {
		private Flusher() {

		}

		@Override
		public void flush(long reference, ReferenceInfo info, boolean closing) throws IOException {
			if (info.getCleanerId() == ERRORED_CLEANER) {
				throw new IOException("Passing a cleaner with the id of ERRORED_CLIENT");
			}
			ByteBuffer data = ByteBuffer.allocate(REF_META_BYTES_COUNT);
			data.putLong(info.getIndex());
			data.putInt(info.getSize());
			data.put(info.getCleanerId());
			data.flip();

			int fileIndex = toFileId(reference);
			long localReference = toLocalReference(fileIndex, reference);
			metaFile.writeAt(fileIndex, localReference * REF_META_BYTES_COUNT, REF_META_BYTES_COUNT, data);
		}

		@Override
		public void flushMultiple(long referenceStart, ReferenceInfo[] infos) throws IOException {
			long[] startLocalReferences = new long[filesCount];
			int[] referencesCount = new int[filesCount];
			Arrays.fill(startLocalReferences, -1);
			ByteBuffer[] buffers = new ByteBuffer[filesCount];
			int maxBufferSize = (infos.length > SEQUENCE_SIZE ? (infos.length / filesCount) + SEQUENCE_SIZE : SEQUENCE_SIZE) * REF_META_BYTES_COUNT;
			for (int i = 0; i < buffers.length; i++) {
				buffers[i] = ByteBuffer.allocate(maxBufferSize);
			}

			for (int delta = 0; delta < infos.length; delta++) {
				ReferenceInfo info = infos[delta];
				long reference = referenceStart + delta;
				int fileId = toFileId(reference);
				long localReference = toLocalReference(fileId, reference);
				if (startLocalReferences[fileId] == -1) {
					startLocalReferences[fileId] = localReference;
				}
				referencesCount[fileId]++;
				if (info.getCleanerId() == ERRORED_CLEANER) {
					throw new IOException("Passing a cleaner with the id of ERRORED_CLIENT");
				}
				if (buffers[fileId].remaining() < REF_META_BYTES_COUNT) {
					throw new IOException();
				}
				buffers[fileId].putLong(info.getIndex());
				buffers[fileId].putInt(info.getSize());
				buffers[fileId].put(info.getCleanerId());
			}
			for (int fileId = 0; fileId < filesCount; fileId++) {
				if (referencesCount[fileId] > 0) {
					buffers[fileId].flip();

					long localPosition = startLocalReferences[fileId] * REF_META_BYTES_COUNT;

					metaFile.writeAt(fileId, localPosition, referencesCount[fileId] * REF_META_BYTES_COUNT, buffers[fileId]);
				}
			}
		}
	}


	private long toReference(int fileId, long localReference) {
		return (fileId + filesCountL * (localReference / SEQUENCE_SIZE_L)) * SEQUENCE_SIZE_L + (localReference % SEQUENCE_SIZE_L);
	}

	private int toFileId(long reference) {
		return (int) ((reference / SEQUENCE_SIZE_L) % filesCountL);
	}

	private long toLocalReference(int fileId, long reference) {
		long chunkId = (reference - (reference % SEQUENCE_SIZE_L)) / SEQUENCE_SIZE_L;
		return (SEQUENCE_SIZE_L * (chunkId - fileId)) / filesCountL + reference % SEQUENCE_SIZE_L;
	}
}
